@@ -1,8 +1,8 @@
 # Automated tests for the `executor' module.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: April 9, 2016
-# URL: https://executor.readthedocs.org
+# Last Change: April 27, 2018
+# URL: https://executor.readthedocs.io
 
 """
 Automated tests for the `executor` package.
@@ -48,23 +48,23 @@ import os
 import pwd
 import random
 import shlex
-import shutil
 import socket
 import sys
 import tempfile
 import time
-import unittest
 import uuid
 
 # External dependencies.
-from humanfriendly import Timer, dedent
-from humanfriendly.compat import StringIO
+from humanfriendly import Timer, compact
+from humanfriendly.testing import TemporaryDirectory, TestCase, retry, run_cli
+from mock import MagicMock
 
 # Modules included in our package.
 from executor import (
+    COMMAND_NOT_FOUND_STATUS,
     DEFAULT_SHELL,
+    DEFAULT_WORKING_DIRECTORY,
     CommandNotFound,
-    ControllableProcess,
     ExternalCommand,
     ExternalCommandFailed,
     execute,
@@ -73,7 +73,16 @@ from executor import (
 )
 from executor.cli import main
 from executor.concurrent import CommandPool, CommandPoolFailed
-from executor.contexts import LocalContext, RemoteContext, create_context
+from executor.contexts import (
+    ChangeRootContext,
+    create_context,
+    LocalContext,
+    RemoteContext,
+    SecureChangeRootContext,
+)
+from executor.process import ProcessTerminationFailed
+from executor.chroot import CHROOT_PROGRAM_NAME
+from executor.schroot import SCHROOT_PROGRAM_NAME
 from executor.ssh.client import (
     DEFAULT_CONNECT_TIMEOUT,
     RemoteCommand,
@@ -87,42 +96,88 @@ from executor.ssh.server import SSHServer
 
 MISSING_COMMAND = 'a-program-name-that-no-one-would-ever-use'
 
+# Initialize a logger for this module.
+logger = logging.getLogger(__name__)
 
-class ExecutorTestCase(unittest.TestCase):
+
+class ExecutorTestCase(TestCase):
 
     """Container for the `executor` test suite."""
 
     def setUp(self):
-        """Set up logging to the terminal and initialize test directories."""
-        # Enable logging to the terminal.
-        try:
-            import coloredlogs
-            coloredlogs.install()
-            coloredlogs.set_level(logging.DEBUG)
-        except ImportError:
-            logging.basicConfig(level=logging.DEBUG)
+        """Set up logging for subprocesses and initialize test directories."""
+        # Set up our superclass.
+        super(ExecutorTestCase, self).setUp()
+        # Enable verbose logging to the terminal for subprocesses.
+        os.environ['COLOREDLOGS_LOG_LEVEL'] = 'DEBUG'
         # Create the directory where superuser privileges are tested.
         self.sudo_enabled_directory = os.path.join(tempfile.gettempdir(), 'executor-test-suite')
         if not os.path.isdir(self.sudo_enabled_directory):
             os.makedirs(self.sudo_enabled_directory)
 
-    def assertRaises(self, type, callable, *args, **kw):
-        """Replacement for :func:`unittest.TestCase.assertRaises()` that returns the exception."""
-        try:
-            callable(*args, **kw)
-        except Exception as e:
-            if isinstance(e, type):
-                # Return the expected exception as a regular return value.
-                return e
-            else:
-                # Don't swallow exceptions we can't handle.
-                raise
-        else:
-            assert False, "Expected an exception to be raised!"
+    def test_graceful_termination(self):
+        """Test graceful termination of processes."""
+        self.check_termination(method='terminate')
 
-    def test_argument_validation(self):
-        """Make sure the external command constructor requires a command argument."""
-        self.assertRaises(TypeError, ExternalCommand)
+    def test_forceful_termination(self):
+        """Test forceful termination of processes."""
+        self.check_termination(method='kill')
+
+    def test_graceful_to_forceful_fallback(self):
+        """Test that graceful termination falls back to forceful termination."""
+        timer = Timer()
+        expected_lifetime = 60
+        with NonGracefulCommand('sleep', str(expected_lifetime), check=False) as cmd:
+            # Request graceful termination even though we know it will fail.
+            cmd.terminate(timeout=1)
+            # Verify that the process terminated even though our graceful
+            # termination request was ignored.
+            assert not cmd.is_running
+            # Verify that the process actually terminated due to the fall back
+            # and not because its expected life time simply ran out.
+            assert timer.elapsed_time < expected_lifetime
+
+    def test_process_termination_failure(self):
+        """Test handling of forceful termination failures."""
+        with NonForcefulCommand('sleep', '60', check=False) as cmd:
+            # Request forceful termination even though we know it will fail.
+            self.assertRaises(ProcessTerminationFailed, cmd.kill, timeout=1)
+            # Verify that the process is indeed still running :-).
+            assert cmd.is_running
+            # Bypass the overrides to get rid of the process.
+            ExternalCommand.terminate_helper(cmd)
+
+    def check_termination(self, method):
+        """Helper method for process termination tests."""
+        with ExternalCommand('sleep', '60', check=False) as cmd:
+            timer = Timer()
+            # We use a positive but very low timeout so that all of the code
+            # involved gets a chance to run, but without slowing us down.
+            getattr(cmd, method)(timeout=0.1)
+            # Gotcha: Call wait() so that the process (our own subprocess) is
+            # reclaimed because until we do so proc.is_running will be True!
+            cmd.wait()
+            # Now we can verify our assertions.
+            assert not cmd.is_running, "Child still running despite graceful termination request!"
+            assert timer.elapsed_time < 10, "It look too long to terminate the child!"
+
+    def test_iterate_buffered(self):
+        """Make sure we can iterate over a command's buffered output."""
+        cmd = ExternalCommand(
+            'for ((i=0; i<10; i++)); do echo $i; done',
+            capture=True, buffered=True,
+        )
+        for i, line in enumerate(cmd):
+            assert i == int(line)
+
+    def test_iterate_unbuffered(self):
+        """Make sure we can iterate over a command's unbuffered output."""
+        cmd = ExternalCommand(
+            'for ((i=0; i<10; i++)); do echo $i; done',
+            capture=True, buffered=False,
+        )
+        for i, line in enumerate(cmd):
+            assert i == int(line)
 
     def test_program_searching(self):
         """Make sure which() works as expected."""
@@ -136,16 +191,19 @@ class ExecutorTestCase(unittest.TestCase):
         # Make sure execute('false') raises an exception.
         self.assertRaises(ExternalCommandFailed, execute, 'false')
         # Make sure execute('exit 42') raises an exception.
-        e = self.assertRaises(ExternalCommandFailed, execute, 'exit 42')
+        shell_cmd = 'echo -n what; echo -n ever; exit 42'
+        e = self.assertRaises(ExternalCommandFailed, execute, shell_cmd, silent=True)
         # Make sure the exception has the expected properties.
-        self.assertEqual(e.command.command_line, ['bash', '-c', 'exit 42'])
+        self.assertEqual(e.command.command_line, ['bash', '-c', shell_cmd])
         self.assertEqual(e.returncode, 42)
+        self.assertTrue('whatever' in e.error_message)
         # Make sure the CommandNotFound exception is raised consistently
         # regardless of the values of the `shell' and `async' options.
         for async in True, False:
             for shell in True, False:
                 cmd = ExternalCommand(MISSING_COMMAND, async=async, shell=shell)
                 self.assertRaises(CommandNotFound, cmd.wait)
+                assert cmd.returncode == COMMAND_NOT_FOUND_STATUS
 
     def test_shell_opt_out(self):
         """Test that callers can always opt out of shell evaluation."""
@@ -166,6 +224,22 @@ class ExecutorTestCase(unittest.TestCase):
         cmd.wait()
         assert cmd.error_type is CommandNotFound
 
+    def test_commands_on_stdin(self):
+        """Test that callers can opt in to shell evaluation for local commands given on standard input."""
+        random_string = uuid.uuid4().hex
+        output = execute(capture=True, shell=True, input='echo %s' % quote(random_string))
+        assert output == random_string
+
+    def test_remote_commands_on_stdin(self):
+        """Test that callers can opt in to shell evaluation for remote commands given on standard input."""
+        random_string = uuid.uuid4().hex
+        with SSHServer() as server:
+            output = remote('127.0.0.1',
+                            capture=True, shell=True,
+                            input='echo %s' % quote(random_string),
+                            **server.client_options)
+            assert output == random_string
+
     def test_stdin(self):
         """Make sure standard input can be provided to external commands."""
         assert execute('tr', 'a-z', 'A-Z', input='test', capture=True) == 'TEST'
@@ -190,6 +264,15 @@ class ExecutorTestCase(unittest.TestCase):
         assert stdout_value in cmd.decoded_stdout
         assert stderr_value in cmd.decoded_stderr
 
+    def test_output_on_error(self):
+        """Check the combination of output capturing and error handling."""
+        for shell in True, False:
+            cmd = ExternalCommand(MISSING_COMMAND, capture=True, check=False, shell=shell)
+            cmd.start()
+            cmd.wait()
+            assert cmd.returncode == COMMAND_NOT_FOUND_STATUS
+            assert cmd.stdout == b''
+
     def test_merged_streams(self):
         """Make sure standard output/error of external commands can be captured together."""
         stdout_value = 'this goes to standard output'
@@ -248,7 +331,7 @@ class ExecutorTestCase(unittest.TestCase):
         """Make sure the standard streams can be redirected to a file and asynchronously stream output to that file."""
         fd, filename = tempfile.mkstemp(prefix='executor-', suffix='-streaming.txt')
         with open(filename, 'w') as handle:
-            cmd = ExternalCommand('for ((i=0; i<25; i++)); do command echo $i; sleep 0.1; done',
+            cmd = ExternalCommand('for ((i=0; i<25; i++)); do echo $i; sleep 0.1; done',
                                   async=True, stdout_file=handle)
             cmd.start()
 
@@ -259,13 +342,6 @@ class ExecutorTestCase(unittest.TestCase):
                 assert len(lines) > 0
                 assert len(lines) < 25
 
-        def expect_most_output():
-            """Expect most but not all output to be readable at some point."""
-            with open(filename) as handle:
-                lines = list(handle)
-                assert len(lines) > 15
-                assert len(lines) < 25
-
         def expect_all_output():
             """Expect all output to be readable at some point."""
             with open(filename) as handle:
@@ -273,8 +349,42 @@ class ExecutorTestCase(unittest.TestCase):
                 assert len(lines) == 25
 
         retry(expect_some_output, 10)
-        retry(expect_most_output, 20)
-        retry(expect_all_output, 30)
+        retry(expect_all_output, 20)
+
+    def test_asynchronous_unbuffered_output(self):
+        """Make sure output buffering to temporary files can be disabled."""
+        cmd = ExternalCommand(
+            *python_golf('import sys',
+                         'sys.stdout.write(sys.stdin.readline().upper())',
+                         'sys.stdout.flush()',
+                         'sys.stdout.write(sys.stdin.readline().upper())'),
+            async=True, buffered=False, capture=True, input=True
+        )
+        with cmd:
+            # Message the command.
+            first_line = 'Hello world?\n'
+            cmd.stdin.write(first_line.lower().encode('ascii'))
+            # Read and check the response.
+            assert cmd.stdout.readline().decode('ascii') == first_line.upper()
+            # Message the command again.
+            second_line = 'Are you still alive?\n'
+            cmd.stdin.write(second_line.lower().encode('ascii'))
+            assert cmd.stdout.readline().decode('ascii') == second_line.upper()
+
+    def test_tty_option(self):
+        """Make sure the ``tty`` option works as expected."""
+        # By default we expect the external command to inherit our standard
+        # input stream (of course this test suite is expected to work
+        # regardless of whether it's connected to a terminal).
+        test_stdin_isatty = python_golf('import sys', 'sys.exit(0 if sys.stdin.isatty() else 1)')
+        assert sys.stdin.isatty() == execute(*test_stdin_isatty, check=False)
+        # If the command's output is being captured then its standard
+        # input stream should be redirected to /dev/null.
+        self.assertRaises(ExternalCommandFailed, execute, *test_stdin_isatty, capture=True)
+        # If the caller explicitly disabled interactive terminal support then
+        # the command's standard input stream should also be redirected to
+        # /dev/null.
+        self.assertRaises(ExternalCommandFailed, execute, *test_stdin_isatty, tty=False)
 
     def test_working_directory(self):
         """Make sure the working directory of external commands can be set."""
@@ -355,16 +465,43 @@ class ExecutorTestCase(unittest.TestCase):
         finally:
             self.assertTrue(execute('rm', '-R', self.sudo_enabled_directory, sudo=True))
 
+    def test_ionice_option(self):
+        """Make sure ``ionice`` can be used."""
+        rsync_command_line = ['rsync', '-a', '/', '/mnt/backups/latest/']
+        expected_ionice_command = ['ionice', '--class', 'idle']
+        command = ExternalCommand(*rsync_command_line, ionice='idle')
+        assert command.ionice == 'idle'
+        print(command.ionice_command)
+        assert command.ionice_command == expected_ionice_command
+        assert command.command_line == (expected_ionice_command + rsync_command_line)
+        self.assertRaises(
+            ValueError,
+            ExternalCommand,
+            'touch', 'something-inappropriate',
+            ionice='unknown-class',
+        )
+
     def test_environment_variable_handling(self):
         """Make sure environment variables can be overridden."""
         # Check that environment variables of the current process are passed on to subprocesses.
-        self.assertEqual(execute('echo $PATH', capture=True), os.environ['PATH'])
+        output = execute('echo $PATH', capture=True)
+        assert output == os.environ['PATH']
         # Test that environment variable overrides can be given to external commands.
-        override_value = str(random.random())
-        self.assertEqual(execute('echo $override',
-                                 capture=True,
-                                 environment=dict(override=override_value)),
-                         override_value)
+        output = execute(
+            'echo $HELLO $WORLD',
+            capture=True,
+            environment=dict(
+                HELLO='Hello',
+                WORLD='world!',
+            ),
+        )
+        assert output == 'Hello world!'
+        # Test that the environment variables of a command can be modified
+        # after the command has been initialized.
+        cmd = ExternalCommand('echo $DELAYED', capture=True)
+        cmd.environment['DELAYED'] = 'Also works fine'
+        cmd.wait()
+        assert cmd.output == 'Also works fine'
 
     def test_simple_async_cmd(self):
         """Make sure commands can be executed asynchronously."""
@@ -430,55 +567,23 @@ class ExecutorTestCase(unittest.TestCase):
         """Callback for :func:`test_callback_evaluation()`."""
         return datetime.datetime.fromtimestamp(float(cmd.output))
 
-    def test_suspend_and_resume_signals(self):
-        """Test the sending of ``SIGSTOP``, ``SIGCONT`` and ``SIGTERM`` signals."""
-        # Spawn a child that will live for a minute.
-        with ExternalCommand('sleep', '60', check=False) as child:
-            # Suspend the execution of the child process using SIGSTOP.
-            child.suspend()
-            # Test that the child process doesn't respond to SIGTERM once suspended.
-            child.terminate(wait=False)
-            assert child.is_running, "Child responded to signal even though it was suspended?!"
-            # Resume the execution of the child process using SIGCONT.
-            child.resume()
-            # Test that the child process responds to signals again after
-            # having been resumed, but give it a moment to terminate
-            # (significantly less time then the process is normally expected
-            # to run, otherwise there's no point in the test below).
-            child.kill(wait=True, timeout=5)
-            assert not child.is_running, "Child didn't respond to signal even though it was resumed?!"
-
-    def test_graceful_command_termination(self):
-        """Test graceful termination of commands."""
-        self.check_command_termination(method='terminate', proxy=False)
-
-    def test_graceful_process_termination(self):
-        """Test graceful termination of processes."""
-        self.check_command_termination(method='terminate', proxy=True)
-
-    def test_forceful_command_termination(self):
-        """Test forceful termination of commands."""
-        self.check_command_termination(method='kill', proxy=False)
-
-    def test_forceful_process_termination(self):
-        """Test forceful termination of commands."""
-        self.check_command_termination(method='kill', proxy=True)
-
-    def check_command_termination(self, method, proxy):
-        """Helper method for command/process termination tests."""
-        with ExternalCommand('sleep', '60', check=False) as cmd:
-            timer = Timer()
-            process = ControllableProcess(pid=cmd.pid) if proxy else cmd
-            # We use a positive but very low timeout so that all of the code
-            # involved gets a chance to run, but without slowing us down.
-            getattr(process, method)(timeout=0.1)
-            # Call ExternalCommand.wait() -> subprocess.Popen.wait() -> os.waitpid()
-            # so that the process (our own subprocess) is reclaimed because
-            # until we do so proc.is_running will be True ...
+    def test_event_callbacks(self):
+        """Make sure the ``start_event`` and ``finish_event`` callbacks are actually invoked."""
+        for async in True, False:
+            results = []
+            cmd = ExternalCommand(
+                'sleep', '0.1',
+                async=async,
+                start_event=lambda cmd: results.append(('started', time.time())),
+                finish_event=lambda cmd: results.append(('finished', time.time())),
+            )
+            cmd.start()
+            mapping = dict(results)
+            assert 'started' in mapping
             cmd.wait()
-            # Now we can verify our assertions.
-            assert not process.is_running, "Child still running despite graceful termination request!"
-            assert timer.elapsed_time < 10, "It look too long to terminate the child!"
+            mapping = dict(results)
+            assert 'finished' in mapping
+            assert mapping['finished'] > mapping['started']
 
     def test_repr(self):
         """Make sure that repr() on external commands gives sane output."""
@@ -599,6 +704,36 @@ class ExecutorTestCase(unittest.TestCase):
                     contents = handle.read()
                 assert filename == ('%s.log' % contents.strip())
 
+    def test_concurrency_control_with_groups(self):
+        """Make sure command pools support ``group_by`` for high level concurrency control."""
+        pool = CommandPool(concurrency=10)
+        for i in range(10):
+            pool.add(ExternalCommand('sleep 0.1', group_by='group-a'))
+        for i in range(10):
+            pool.add(ExternalCommand('sleep 0.1', group_by='group-b'))
+        while not pool.is_finished:
+            pool.spawn()
+            # Make sure we never see more than two commands running at the same
+            # time (because the commands are spread over two command groups).
+            assert pool.num_running <= 2
+            pool.collect()
+
+    def test_concurrency_control_with_dependencies(self):
+        """Make sure command pools support ``dependencies`` for low level concurrency control."""
+        pool = CommandPool(concurrency=10)
+        group_one = [ExternalCommand('sleep 0.1') for i in range(5)]
+        group_two = [ExternalCommand('sleep 0.1', dependencies=group_one) for i in range(5)]
+        group_three = [ExternalCommand('sleep 0.1', dependencies=group_two) for i in range(5)]
+        for group in group_one, group_two, group_three:
+            for cmd in group:
+                pool.add(cmd)
+        while not pool.is_finished:
+            pool.spawn()
+            # Make sure we never see more than one group of commands running at
+            # the same time (because we've set up the dependencies like this).
+            assert pool.num_running <= 5
+            pool.collect()
+
     def test_ssh_user_at_host(self):
         """Make sure a username can be injected via an SSH alias."""
         cmd = RemoteCommand('root@host', 'true')
@@ -709,6 +844,63 @@ class ExecutorTestCase(unittest.TestCase):
             assert len(log_files) == len(ssh_aliases)
             assert all(os.path.getsize(os.path.join(directory, fn)) > 0 for fn in log_files)
 
+    def test_chroot_command(self):
+        """
+        Test support for chroot commands.
+
+        For now this test doesn't actually run ``chroot`` because automating
+        the creation of chroots using ``debootstrap`` just to run these tests
+        is a lot of work that I haven't done (yet).
+        """
+        chroot = '/var/lib/chroots/executor'
+        chroot_group = 'my-group'
+        chroot_user = 'my-user'
+        command = ['echo', '42']
+        context = ChangeRootContext(chroot)
+        cmd = context.prepare(*command, chroot_group=chroot_group, chroot_user=chroot_user)
+        assert CHROOT_PROGRAM_NAME in cmd.command_line
+        assert '--userspec=%s:%s' % (chroot_user, chroot_group) in cmd.command_line
+        assert cmd.command_line[-len(command):] == command
+        # Make sure sudo is avoided when possible.
+        cmd = context.prepare('apt-get', 'update', sudo=True)
+        command_in_chroot = cmd.command_line[cmd.command_line.index('chroot'):]
+        assert 'sudo' not in command_in_chroot
+        # Make sure sudo is used when necessary.
+        cmd = context.prepare('apt-get', 'update', chroot_user='nobody', sudo=True)
+        command_in_chroot = cmd.command_line[cmd.command_line.index('chroot'):]
+        assert 'sudo' in command_in_chroot
+        # Make sure the working directory is handled correctly.
+        directory_in_chroot = '/relative/to/chroot'
+        cmd = context.prepare('pwd', directory=directory_in_chroot)
+        assert cmd.directory == DEFAULT_WORKING_DIRECTORY
+        assert cmd.chroot_directory == directory_in_chroot
+        assert any(directory_in_chroot in token for token in cmd.command_line)
+
+    def test_schroot_command(self):
+        """
+        Test support for schroot commands.
+
+        For now this test doesn't actually run ``schroot`` because automating
+        the installation of ``schroot`` and the creation of chroots using
+        ``debootstrap`` just to run these tests is a lot of work that I haven't
+        done (yet).
+        """
+        chroot_name = 'name-of-chroot'
+        chroot_user = 'user-in-chroot'
+        chroot_directory = '/path/relative/to/chroot'
+        command = ['echo', '42']
+        context = SecureChangeRootContext(chroot_name, chroot_directory=chroot_directory, chroot_user=chroot_user)
+        cmd = context.prepare(*command)
+        assert SCHROOT_PROGRAM_NAME in cmd.command_line
+        assert ('--chroot=%s' % chroot_name) in cmd.command_line
+        assert ('--user=%s' % chroot_user) in cmd.command_line
+        assert ('--directory=%s' % chroot_directory) in cmd.command_line
+        assert cmd.command_line[-len(command):] == command
+        other_chroot_directory = '/other/path/relative/to/chroot'
+        cmd.directory = other_chroot_directory
+        assert cmd.directory == DEFAULT_WORKING_DIRECTORY
+        assert cmd.chroot_directory == other_chroot_directory
+
     def test_create_context(self):
         """Test context creation."""
         assert isinstance(create_context(), LocalContext)
@@ -718,6 +910,24 @@ class ExecutorTestCase(unittest.TestCase):
         assert create_context(sudo=True).options['sudo'] is True
         assert create_context(sudo=False).options['sudo'] is False
 
+    def test_lsb_release_shortcuts(self):
+        """Test the ``lsb_release`` shortcuts."""
+        try:
+            # The following tests should pass on my laptops and Travis CI.
+            context = LocalContext()
+            assert context.distributor_id == 'ubuntu'
+            assert context.distribution_codename in ('precise', 'trusty', 'xenial')
+        except AssertionError:
+            # But I don't want this test to fail on `unexpected'
+            # platforms so here's a pragmatic compromise :-).
+            return self.skipTest("assuming unsupported platform")
+
+    def test_lsb_release_error_handling(self):
+        """Test that the ``lsb_release`` shortcuts don't raise exceptions on unsupported platforms."""
+        context = LocalContext(environment=dict(PATH=''))
+        assert context.distributor_id == ''
+        assert context.distribution_codename == ''
+
     def test_local_context(self):
         """Test a local command context."""
         self.check_context(LocalContext())
@@ -735,7 +945,11 @@ class ExecutorTestCase(unittest.TestCase):
         assert context.cpu_count >= 1
         # Test context.execute() and cleanup().
         random_file = os.path.join(self.sudo_enabled_directory, uuid.uuid4().hex)
+        # Prepare to test context.cleanup() with a callable.
+        cleanup_callback = MagicMock()
         with context:
+            # Schedule to invoke our callback before the with block ends.
+            context.cleanup(cleanup_callback, 42, keyword='value')
             # Make sure the test directory exists.
             assert context.exists(self.sudo_enabled_directory)
             assert context.is_directory(self.sudo_enabled_directory)
@@ -751,10 +965,19 @@ class ExecutorTestCase(unittest.TestCase):
             # Make sure the file is readable and writable.
             assert context.is_readable(random_file)
             assert context.is_writable(random_file)
+            # Make sure the file isn't executable.
+            context.execute('chmod', '-x', random_file)
+            assert not context.is_executable(random_file)
+            # Make sure we can make the file executable.
+            context.execute('chmod', '+x', random_file)
+            assert context.is_executable(random_file)
             # Schedule to clean up the file.
             context.cleanup('rm', '-f', random_file)
             # Make sure the file hasn't actually been removed yet.
             assert context.exists(random_file)
+            # Find the file using a filename pattern.
+            matches = context.glob(os.path.join(self.sudo_enabled_directory, '*'))
+            assert random_file in matches
             # The following tests only make sense when we're not already
             # running with superuser privileges.
             if os.getuid() != 0:
@@ -764,6 +987,8 @@ class ExecutorTestCase(unittest.TestCase):
                 # Make sure the file is no longer readable or writable.
                 assert not context.is_readable(random_file)
                 assert not context.is_writable(random_file)
+        # Make sure our cleanup callback was invoked correctly.
+        assert cleanup_callback.called_with(42, keyword='value')
         # Make sure the file has been removed (__exit__).
         assert not context.exists(random_file)
         # Test context.capture().
@@ -779,8 +1004,34 @@ class ExecutorTestCase(unittest.TestCase):
         # Make sure the contents are correct.
         actual_contents = context.read_file(random_file)
         assert actual_contents == expected_contents
-        # Test context.list_entries() and make sure it doesn't mangle filenames
-        # containing whitespace.
+        # Test the happy path in context.atomic_write().
+        random_file = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
+        expected_contents = bytes(random.randint(0, 255) for i in range(25))
+        assert not context.exists(random_file)
+        with context.atomic_write(random_file) as temporary_file:
+            context.write_file(temporary_file, expected_contents)
+            assert not context.exists(random_file)
+        assert not context.exists(temporary_file)
+        assert context.exists(random_file)
+        assert context.read_file(random_file) == expected_contents
+        # Test the failure handling in context.atomic_write().
+        random_file = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
+        try:
+            assert not context.exists(random_file)
+            with context.atomic_write(random_file) as temporary_file:
+                context.write_file(temporary_file, '')
+                assert context.exists(temporary_file)
+                # Interrupt the `with' block by raising an exception.
+                raise Exception
+        except Exception:
+            pass
+        finally:
+            # Make sure the temporary file was cleaned up.
+            assert not context.exists(temporary_file)
+            # Make sure the target file wasn't created.
+            assert not context.exists(random_file)
+        # Test context.list_entries() and make sure it doesn't
+        # mangle filenames containing whitespace.
         nasty_filenames = [
             'something-innocent',
             'now with spaces',
@@ -800,33 +1051,50 @@ class ExecutorTestCase(unittest.TestCase):
     def test_cli_usage(self):
         """Make sure the command line interface properly presents its usage message."""
         for arguments in [], ['-h'], ['--help']:
-            with CaptureOutput() as stream:
-                assert run_cli(*arguments) == 0
-                assert "Usage: executor" in str(stream)
+            returncode, output = run_cli(main, *arguments)
+            assert returncode == 0
+            assert "Usage: executor" in output
 
     def test_cli_return_codes(self):
         """Make sure the command line interface doesn't swallow exit codes."""
-        assert run_cli(*python_golf('import sys; sys.exit(0)')) == 0
-        assert run_cli(*python_golf('import sys; sys.exit(1)')) == 1
-        assert run_cli(*python_golf('import sys; sys.exit(42)')) == 42
+        returncode, output = run_cli(main, *python_golf('import sys', 'sys.exit(0)'))
+        assert returncode == 0
+        returncode, output = run_cli(main, *python_golf('import sys', 'sys.exit(1)'))
+        assert returncode == 1
+        returncode, output = run_cli(main, *python_golf('import sys', 'sys.exit(42)'))
+        assert returncode == 42
 
     def test_cli_fudge_factor(self, fudge_factor=5):
         """Try to ensure that the fudge factor applies (a bit tricky to get right) ..."""
         def fudge_factor_hammer():
             timer = Timer()
-            assert run_cli('--fudge-factor=%i' % fudge_factor, *python_golf('import sys; sys.exit(0)')) == 0
+            returncode, output = run_cli(
+                main,
+                '--fudge-factor=%i' % fudge_factor,
+                *python_golf('import sys', 'sys.exit(0)')
+            )
+            assert returncode == 0
             assert timer.elapsed_time > (fudge_factor / 2.0)
         retry(fudge_factor_hammer, 60)
 
     def test_cli_exclusive_locking(self):
         """Ensure that exclusive locking works as expected."""
-        run_cli('--exclusive', *python_golf('import sys; sys.exit(0)')) == 0
+        returncode, output = run_cli(
+            main,
+            '--exclusive',
+            *python_golf('import sys', 'sys.exit(0)')
+        )
+        assert returncode == 0
 
     def test_cli_timeout(self):
         """Ensure that external commands can be timed out."""
         def timeout_hammer():
             timer = Timer()
-            assert run_cli('--timeout=5', *python_golf('import time; time.sleep(10)')) != 0
+            returncode, output = run_cli(
+                main, '--timeout=5',
+                *python_golf('import time', 'time.sleep(10)')
+            )
+            assert returncode != 0
             assert timer.elapsed_time < 10
         retry(timeout_hammer, 60)
 
@@ -846,90 +1114,32 @@ def tokenize_command_line(cmd):
     return sum(map(shlex.split, cmd.command_line), [])
 
 
-def retry(func, timeout):
-    """Retry a function until it no longer raises assertion errors or time runs out before then."""
-    time_started = time.time()
-    while True:
-        timeout_expired = (time.time() - time_started) >= timeout
-        try:
-            return func()
-        except AssertionError:
-            if timeout_expired:
-                raise
-
-
-def python_golf(statements):
+def python_golf(*statements):
     """Generate a Python command line."""
-    return sys.executable, '-c', dedent(statements)
+    return sys.executable, '-c', '; '.join(statements)
 
 
-def run_cli(*arguments):
-    """Run the command line interface (in the same process)."""
-    saved_argv = sys.argv
-    try:
-        sys.argv = ['executor'] + list(arguments)
-        main()
-    except SystemExit as e:
-        return e.code
-    else:
-        return 0
-    finally:
-        sys.argv = saved_argv
+class NonGracefulCommand(ExternalCommand):
+
+    """Wrapper for :class:`~executor.process.ControllableProcess` that disables graceful termination."""
+
+    def terminate_helper(self, *args, **kw):
+        """Swallow graceful termination signals."""
+        self.logger.debug(compact("""
+            Process termination using subprocess.Popen.terminate()
+            intentionally disabled to simulate processes that refuse to
+            terminate gracefully ..
+        """))
 
 
-class CaptureOutput(object):
+class NonForcefulCommand(NonGracefulCommand):
 
-    """Context manager that captures what's written to :data:`sys.stdout`."""
+    """Wrapper for :class:`~executor.process.ControllableProcess` that disables graceful and forceful termination."""
 
-    def __init__(self):
-        """Initialize a string IO object to be used as :data:`sys.stdout`."""
-        self.stream = StringIO()
-
-    def __enter__(self):
-        """Start capturing what's written to :data:`sys.stdout`."""
-        self.original_stdout = sys.stdout
-        sys.stdout = self.stream
-        return self
-
-    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
-        """Stop capturing what's written to :data:`sys.stdout`."""
-        sys.stdout = self.original_stdout
-
-    def __str__(self):
-        """Get the text written to :data:`sys.stdout`."""
-        return self.stream.getvalue()
-
-
-class TemporaryDirectory(object):
-
-    """
-    Easy temporary directory creation & cleanup using the :keyword:`with` statement.
-
-    Here's an example of how to use this:
-
-    .. code-block:: python
-
-       with TemporaryDirectory() as directory:
-           # Do something useful here.
-           assert os.path.isdir(directory)
-    """
-
-    def __init__(self, **options):
-        """
-        Initialize a :class:`TemporaryDirectory` object.
-
-        :param options: Any keyword arguments are passed on to :func:`tempfile.mkdtemp()`.
-        """
-        self.options = options
-        self.temporary_directory = None
-
-    def __enter__(self):
-        """Create the temporary directory."""
-        self.temporary_directory = tempfile.mkdtemp(**self.options)
-        return self.temporary_directory
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        """Destroy the temporary directory."""
-        if self.temporary_directory is not None:
-            shutil.rmtree(self.temporary_directory)
-            self.temporary_directory = None
+    def kill_helper(self, *args, **kw):
+        """Swallow forceful termination signals."""
+        self.logger.debug(compact("""
+            Process termination using subprocess.Popen.kill() intentionally
+            disabled to simulate processes that refuse to terminate
+            forcefully ..
+        """))
